/*
 * $Id: BodyPart.java,v 1.3 1996/01/19 01:51:06 iansmith Exp $
 * $Author: iansmith $
 * $Log: BodyPart.java,v $
 * Revision 1.3  1996/01/19  01:51:06  iansmith
 * Documentation improvements in the javadoc comments.
 *
 * Revision 1.2  1996/01/15  18:43:56  iansmith
 * Add a call to get the IMAP part number of a BodyPart. Probably not
 * generally useful.
 *
 * Added a new command called PARTIAL() which is a version of FETCH()
 * that only grabs part of a message's data. As of now, this is untested.
 *
 * Added code in the getData() routine so that calling this function now
 * has the implication of calling nowSeen() on the message, marking it as
 * seen and (possibly) not new anymore.
 *
 * Revision 1.1  1996/01/10  21:49:38  iansmith
 * Initial revision
 *
 *
 */

package GIT.GVU.IMAP;
import java.lang.*;
import java.util.*;

/**
 * Class to hold the information about a single part of an email message.
 * A message with multiple parts (i.e. attachments) will have multiple
 * BodyParts.
 */
public class BodyPart {
  /**
   * Construct a body part object.
   */
  BodyPart(String t, String sub, Vector a, String i,
           String d, String e, int si, int l, String p, Message m) {
    type=t;
    subtype=sub;
    /* attributes may be null, but the calling code does this one */
    attributes=a;
    /* might be no id... */
    if (i.equals("NIL")==true) {
      id=null;
    } else {
      id=i;
    }
    /* might have no description */
    if (d.equals("NIL")==true) {
      description=null;
    } else {
      description=d;
    }
    /* might not have an encoding if its multipart */
    if (e.equals("NIL")==true) {
      encoding=null;
    } else {
      encoding=e;
    }
    /* always has these two but the lines might not mean anything */
    size=si;
    lines=l;
    /* default to being a smiple guy */
    firstChild=null;
    nextSibling=null;
    /* set the part number */
    partNumber=p;
    /* message */
    message=m;
    /* default to no embedded message */
    embeddedMessage=null;
    /* data is not here yet */
    data=null;
  }
  /**
   * This variable holds all list of all the children of this BodyPart.
   * Example: If this body part is Multipart/Mixed then it will have
   * children, and they will all be siblings of each other, organized
   * as a list.
   */
  BodyPart firstChild;
  /**
   * Siblings of this BodyPart.
   */
  BodyPart nextSibling;
  /**
   * This is the MIME type of the bodypart 
   */
  String type;
  /**
   * This is the MIME subtype of the bodypart
   */
  String subtype;
  /**
   * This is a vector of strings. This is a set of attribue-value pairs
   * which describes this BodyPart of this message.
   * I.e. You will get a vector of "CHARSET" and "us-ascii" in most cases.
   */
  Vector attributes;
  /**
   * MIME part ID
   */
  String id;
  /**
   * Description (english text) about this part of the message
   */
  String description;
  /**
   * Encoding of the message 
   */
  String encoding;
  /**
   * size of the message in bytes before decoding (the size you get)
   * Note: Will be -1 if a multipart message.
   */
  int size ;
  /**
   * size of the message in lines  (if applicable)
   */
  int lines;
  /**
   * This string holds the part number of the body part for use with
   * the FETCH() command.
   */ 
  String partNumber;
  /**
   * This is the message that this body part is associated with.
   */
  Message message;
  /**
   * Get the first child of this BodyPart. Note: If this part has no children
   * (i.e.  its not a multipart or message type) this func returns null.
   *
   * @return first child of this message (BodyPart)
   */
  public BodyPart getFirstChild() { return firstChild;};
  /**
   * Get the next sibling of this body part. This (obviously) will return
   * no if the message has no siblings OR if this message is the last 
   * of a set of siblings.
   *
   * @return next sibling of this message (BodyPart)
   */
  public BodyPart getNextSibling() { return nextSibling;};
  /**
   * Get the MIME type of this message as a string.
   * 
   * @return a string with the MIME type of this message
   */
  public String getMIMEType() { return type;};
  /**
   * Get the MIME subtype of this message as a string.
   * 
   * @return a string with the MIME subtype of this message
   */
  public String getMIMESubType() { return subtype;};
  /**
   * Get a vector of attributes about the body part. Note: This might
   * be null if none were supplied. Generally, this is going to have
   * at least the pair CHARSET <somecharset>. The contents of 
   * this vector are strings in Attribute, Value order (i.e. evens are
   * names of attributes, odds are their values).
   * 
   * @return a vector of strings which is attribute-value pairs about this message
   */
  public Vector getAttributes() { return attributes;};
  /**
   * Get the MIME identifier of this body part. If none was supplied, this
   * might be null.   
   * 
   * @return a string with this messages MIME identifier or null
   */
  public String getIdentifier() { return id; };
  /**
   * Get the descriptive string about this bodypart.  Note: Might be null.
   *
   * @return a string with a description of this body part (or null)
   */
  public String getDescription() { return description;};
  /**
   * Get the encoding of this body part. XXX This should probably be
   * an enumeration but it returns a string XXX  Note: The multipart
   * body part of a multipart message has no encoding.
   *
   * @return a string with the encoding type of this body part
   */
  public String getEncoding() { return encoding; };
  /**
   * Get the number of bytes in this body part. This is the size before any
   * decoding you do (i.e. the transmitted size).  Note: A multipart BodyPart
   * has no size; this function will return -1.
   * 
   * @return an integer with the size of this body part
   */
  int getSize() { return size;};
  /**
   * Get the number of lines of this body part. This will be -1 for all body
   * parts which don't have TEXT as their MIME type.
   *
   * @return an int with the number of lines in this message (or -1)
   */
  int getLines() { return lines;};
  /**
   * Set the child field. The object passed should be the first child of this object.
   *
   * @param BodyPart c childObject
   */
  void setFirstChild(BodyPart b) { firstChild=b;};
  /**
   * Set the nextSibling field. 
   *
   * @param BodyPart b becomes the sibling of this object
   */
  void setSibling(BodyPart b) { nextSibling=b;};
  /**
   * Grab the data for this part from the server via the FETCH message.
   * Note this is a call to "FETCH n BODY[p]" where n is the message number
   * and p is the part number.  Also note that this is cached, so repeated
   * calls to this method have no effect after the first one. Thus, don't
   * hesitate to call this before calling getData(). Also note: this call modifies
   * the \Seen flag on the message if its not already set via nowSeen().
   *  
   *
   * @exception IMAPException is propagated
   * @exception InternalErr is propagated 
   * @return  true if this succeeded, false otherwise
   */
  public boolean FETCH() throws IMAPException, InternalErr {
    Mailbox box;
    Protocol p;
    Vector result=new Vector();
    boolean worked;

    /* caching: if we already have the data , don't bother */
    if (data!=null) {
      return true;
    }
    /* get the mailbox we are in */
    box=message.getMailbox();
    /* get the protocol object */
    p=box.getProtocol();
    /* send the message to the server */
    p.getIO().roundTrip("FETCH " + message.getNumber() +  
                        " BODY[" + getPartNumber()+"]" ,result);
    /* now tell the mailbox about us */
    box.setBodyPartReceiver(this);
    /* dispatch the results */
    p.dispatchVector(result,false);
    /* we got our message, don't need it anymore */
    box.setBodyPartReceiver(null);
    /* did it work ?*/
    worked=(Util.replyToken(result,"OK")==true);
    /* if so, update the flags on Mailbox */
    if (worked) {
      /* tell it we have seen this message */
      message.nowSeen();
    }
    return (worked);
  }
  /**
   * This call is identical to the FETCH() command but only grabs part of the data. 
   * It resets this body parts data to be the result of this command. Thus, to get all
   * the data, you must issue multiple calls to PARTIAL and getData(). This command
   * would be a good idea if you think (look at getSize()!) the data you are going to
   * retreive is very large. Using PARTIAL you can get it in pieces. Note: if you need
   * to base64 decode this data, be sure you don't end up calling decodeUnit() with
   * a number of bytes that is not a multiple of 4!
   * 
   * @param int first offset of first byte to retreive
   * @param int num number of bytes to retreive 
   * @exception IMAPException is propagated
   * @exception InternalErr is propagated 
   * @return true if the PARTIAL command was successful.
   */
  public boolean PARTIAL(int first, int num) throws InternalErr, IMAPException{
    Integer f=new Integer(first),n=new Integer(num);
    Mailbox box;
    Protocol p;
    Vector result=new Vector();
    boolean worked;

    /* get the mailbox we are in */
    box=message.getMailbox();
    /* get the protocol object */
    p=box.getProtocol();
    /* send the message to the server */
    p.getIO().roundTrip("PARTIAL " + message.getNumber() +  
                        " BODY[" + getPartNumber()+"]" + " " +  
                        f.toString() + " " + n.toString(),result);
    /* now tell the mailbox about us */
    box.setBodyPartReceiver(this);
    /* dispatch the results */
    p.dispatchVector(result,false);
    /* we got our message, don't need it anymore */
    box.setBodyPartReceiver(null);
    /* did it work ?*/
    worked=(Util.replyToken(result,"OK")==true);
    /* if so, update the flags on Mailbox */
    if (worked) {
      /* tell it we have seen this message */
      message.nowSeen();
    }
    return (worked);
  }
  /**
   * Gets the IMAP part number of a message. 
   * 
   *  Probably only useful for people working on the protocol level.
   *
   * MULITPART and MESSAGE/RFC822 have no part number.
   *
   * @return String with the printed form of the message's part number 
   */
  public String getPartNumber() { return partNumber;};
  /**
   * This function gets the Message object that is embedded in this object,
   * if it has one. This will be null if this BodyPart is not of type MESSAGE
   * and subtype RFC822.
   * 
   * @return a Message that is held by this body part or null.
   */
  public Message getEmbeddedMessage() { return embeddedMessage;};
  /**
   * This object holds the embedded message for MESSAGE/RFC822 body parts.
   */
  Message embeddedMessage;
  /**
   * Set the embedded message of an body part.
   * @param Message m message to embed
   */
  void setEmbeddedMessage(Message m) { embeddedMessage=m;};
  /**
   * This function gets called by the mailbox when it finds a fetch message
   * that we are interested in.
   */
  void handleFetch(String s)  throws InternalErr {
    int index;
    Vector result;

    /* find first ( */
    index=s.indexOf("(");
    /* parse! */
    result=Util.buildList(s,index);
    /* remove excess parsing data */
    result=(Vector)Util.removeParsingData(result);
    /* now result slot 1 should be our text ... sanity check*/
    if (result.elementAt(1) instanceof String==false) {
      throw new InternalErr("Unable to find the text of this body part!");
    }
    data=(String)result.elementAt(1);
  }
  /** 
   * Retreive the text of this body part as a string. Note: This information is cached,
   * so retreiving it repeatedly gives the same value.
   *
   * @return String with the text of this body part 
   */
  public String getData() { return data;};
  /**
   * This variable holds the value of this body part's data
   */
  String data;
  
}
