/*
 *  $Id: Mailbox.java,v 1.6 1996/01/19 01:55:32 iansmith Exp $
 *  $Author: iansmith $
 *  $Log: Mailbox.java,v $
 * Revision 1.6  1996/01/19  01:55:32  iansmith
 * Added support so a mailbox knows what protcol object it is associated with.
 *
 * Added support for Asynchronous Actions at EXPUNGE and RECENT
 * time. This included both using the AA objects but also setting them.
 *
 * Major documentation improvements in the javadoc comments.
 *
 * Added the COPY function().
 *
 * Added the CHECK() function.
 *
 * Revision 1.5  1996/01/15  18:48:59  iansmith
 * Changed this code to reflect that some of the MIME decoding code is
 * now in the MIME object, not the Util object as it was before.
 *
 * Added some code in the handleFetch() routine to deal with simple FETCH
 * FLAGS responses. These can/will result as a response to a STORE command.
 *
 * Added a new version of FETCH() which takes an integer instead of a string.
 *
 * Revision 1.4  1996/01/10  21:57:01  iansmith
 * Added support for a "receiver" which is a BodyPart which installs itself
 * into this object via setBodyPartReceiver() and it 'intercepts' FETCH
 * messages that arrive at this object. This allows us to have the user
 * do FETCH() on BodyParts and have it "do the right thing."  This was
 * accomplished by using the walkBackward() method of this object.
 *
 * All of the code to create a message was put into a function createMessage
 * which will build a message given the Vectors that were parsed out of
 * the FETCH response. This is useful as this createMessage() code is also
 * used by the Util class to build encapsulated messages inside RFC822
 * messages. This change basically amounted to moving a bunch of code that
 * was in handleFetch() into createMessage() and adding a couple of
 * parameters that weren't needed at that time. The main new parameter is
 * the body "basePart" which is used for naming MIME parts. This needs
 * to be a parameter so the Util code can name its "encapsulated" MIME
 * parts as subparts of an RFC822 message.
 *
 * Added caching to the FETCH() message so it doesn't do anything now if
 * the message is already in cache.
 *
 * Revision 1.3  1996/01/08  06:00:08  iansmith
 * Modified the mailbox object to have a messageCache which is a
 * hashtable which keeps track of the mapping from message numbers
 * to Message objects. It is added to by the FETCH() API call and
 * is accessed via getMessage().
 *
 * Modified some API functions to throw InternalErr as the
 * IO.roundTrip() call now throws it and I wanted to pass it through.
 *
 * Modified the walkForward() code to now accept FETCH as one of the
 * messages it dispatches.
 *
 * Added the handleFecth() routine which can parse the header
 * information retreived by the FETCH() routine.
 *
 * Made the FETCH() routine now be a call to FETCH (FLAGS RFC822.SIZE
 * ENVELOPE BODYSTRUCTURE) so we can get everything we need in just
 * one mail message.
 *
 * Added the getMessage() routine to access the message cache.
 *
 * Added the flushCache() routine to flush the cache. Currently unused.
 *
 *  Revision 1.2  1996/01/06 21:22:15  iansmith
 *  Additon of part of the FETCH() function.
 *
 *  Revision 1.1  1996/01/06 19:50:46  iansmith
 *  Initial revision
 *
 */
package GIT.GVU.IMAP;
import java.lang.*;
import java.util.*;

/**
 * Class to hold informationa about a selected mailbox.
 *
 * @version $Id: Mailbox.java,v 1.6 1996/01/19 01:55:32 iansmith Exp $
 * @author Ian Smith
 */
public class Mailbox {
  /** This function tells you if this mailbox is readonly or not.
   * Note: see RFC1730 for some cases where you can modify the per-use
   * state of read-only mailboxes when then are accessed via SELECT.
   * @return true if the mailbox is read-only, false otherwise
   */
  public boolean isReadOnly() {return readonly;};
  /** Construct a new mailbox object. 
   *
   * @param r boolean for readonly mailbox
   * @param p protocol object which we are connected to
   */
  Mailbox(boolean r, Connection c) {
    readonly=r;
    recent=-1; /* user's shouldn't see this as -1 */
    exists=-1; /* user's shouldn't see this as -1 */ 
    flags=new Vector(); /* this will be holding the flags as strings */
    permanentFlags=new Vector(); /* this will hold the permanent flags */
    /* set up the cache */
    messageCache=new Hashtable(100,0.9f);
    /* initially no receiver */
    receiver=null;
    /* we need to know our connection in case they close the box */
    connection=c;
    /* initially we don't have any action objects */
    expungeAction=null;
    recentAction=null;
  }
  /** Variable to hold our readonly state.
   */
  public boolean readonly;
 /** 
   * Forward protocol dispatching 
   * 
   * @param s the string to dispatch 
   * @exception InternalErr is propagated
   * @return true if we dispatched the string or someone downstream did, otherwise false
   */
  boolean walkForward(String s) throws InternalErr{
    StringTokenizer st;
    String tmp;
    int num=0; /* must assign something to it to make compiler happy */
    boolean hadNumber=true;

    /* look for the flags message */
    if (Util.lookingAt(s,"* FLAGS")==true) {
      handleFlags(s);
      return true;
    }
    /* look for the permanent flags message */
    if (Util.lookingAt(s,"* OK [PERMANENTFLAGS")==true) {
      handlePermanentFlags(s);
      return true;
    }
    /* look at the 3rd token of a message starting with * to find the */
    /* EXISTS, RECENT, & FETCH  messages */
    st=new StringTokenizer(s);
    /* although EXISTS and RECENT have exactly 3 tokens, FETCH could have */
    /* any number so we just test for > 2 */
    if (st.countTokens()>2) {
      st.nextToken(); /* get rid of the "*" */

      /* try to get a number from the second token */
      try {
	num=Integer.parseInt(st.nextToken(),10);
      } catch (NumberFormatException n) {
	hadNumber=false;
      }
      
      /* did we get a number? */
      if (hadNumber==true) {
	/* we got a number ... look at the 3rd token */
	tmp=st.nextToken();
	/* got it, check for our friends */
	if (tmp.equals("EXISTS")==true) {
	  handleExists(s,num);
	  return true;
	}
	if (tmp.equals("RECENT")==true){
	  handleRecent(s,num);
	  return true;
	}
	if (tmp.equals("FETCH")==true) {
	  handleFetch(s,num);
	  return true;
	}
	if (tmp.equals("EXPUNGE")==true) {
	  handleExpunge(s,num);
	  return true;
	}
      }
    }
    /* we didn't handle it, tell the caller */
    return false;
  }
  /** 
   * Backward protocol dispatching 
   * 
   * @param s the string to dispatch 
   * @exception InternalErr is propagated
   * @return true if we dispatched the string or someone downstream did, otherwise false
   */
  boolean walkBackward(String s) throws InternalErr {
    StringTokenizer st;
    int num;
    boolean hadNumber=true;
    /* Ok, we do the trickery here to allow a body part to get the FETCH result */
    /* if we have a receiver object currently, tell it about fetches */
    if (receiver!=null) {
      /* ok, try to process this string as a fetch */
      st=new StringTokenizer(s);
      /* must have enough tokens */
      if (st.countTokens()>2) {
	/* try for * */
	if (st.nextToken().equals("*")==true) {
	  /* try to get a number from the second token */
	  try {
	    num=Integer.parseInt(st.nextToken(),10);
	  } catch (NumberFormatException n) {
	    hadNumber=false;
	  }
	}
	/* got a number? */
	if (hadNumber==true) {
	  /* is this a fetch */
	  if (st.nextToken().equals("FETCH")==true) {
	    /* ok, got it, now hand it to the BodyPart */
	    receiver.handleFetch(s);
	    return true;
	  }
	}
      }
    }
    /* end of local processing */
    return false;
  }
  /**
   * Code to deal with the untagged FLAGS message from the server.
   * 
   * @param s string with the server message
   */
  void handleFlags(String s) {
    StringTokenizer st;
    String r;

    st=new StringTokenizer(s);
    st.nextToken(); /* skip the asterisk */
    st.nextToken(); /* skip the FLAGS */
    /* now, grab the remainder of the string up to the final ')' */
    r=st.nextToken(")");
    /* assign the flags local var to be the value of this parse */
    flags=Util.parseStringList(r);
  }
  /** 
   * Code to deal with the untagged RECENT message from the server
   *
   * @param s string with the server message
   * @param n int with the parameter of the RECENT message already parsed out
   */
  void handleRecent(String s, int n) {
    /* ignore s, we don't need it */
    recent=n;
    /* if we have a recent action, deal with it */
    if (recentAction!=null) {
      recentAction.run("RECENT",new Integer(n));
    }
  }
  /** 
   * Code to deal with the untagged EXISTS message from the server
   *
   * @param s string with the server message
   * @param n int with the parameter of the RECENT message already parsed out
   */
  void handleExists(String s, int n) {
    /* ignore s, we don't need it */
    exists=n;
  }
  /**
   * This is an internal use function which converts some flags, an envelope
   * vector, bodystructure, and a size into a Message. The final parameter is 
   * used to initialize the part numbers of the body parts. Note: This method
   * gets called in the simple case to build a message from a fetched message
   * structure, but <EM>also</EM> gets called from the structureToPart() 
   * routine over in the Util object. This is because of the possibility of 
   * encapsulated RFC822 messages. 
   * 
   * @param Vector flags flags (may be null)
   * @param Vector envelope envelope structure of the message
   * @param Vector body bodystructure from the message.
   * @param int size size of the message in bytes
   * @param Mailbox m the mailbox to associate with this message
   * @param String basePart the base part of the message's body part's mime parts
   * 
   * @exception InternalErr  is thrown if we encounter bogus values in the BCC or CC field
   */
  static Message createMessage(Vector flags, Vector envelope, Vector body, 
                       int size, Mailbox m, String basePart, int msgno) 
                       throws InternalErr{
    Message message;
    String date, subject, irt, mid, tmp;
    Vector from, sender, replyTo, to, cc, bcc;
    BodyPart bp;

    /* of the envelope, slot 0 is the date */
    date=(String)envelope.elementAt(0);
    /* subject is slot 1 of envelope */
    subject=(String)envelope.elementAt(1);
    /* now we need the from line ... its slot 2, and its a vector */
    from=Util.parseAddresses((Vector)envelope.elementAt(2));
    /* slot 3=sender */
    sender=Util.parseAddresses((Vector)envelope.elementAt(3));
    /* slot 4=reply to */
    replyTo=Util.parseAddresses((Vector)envelope.elementAt(4));
    /* slot 5=to */
    to=Util.parseAddresses((Vector)envelope.elementAt(5));
    /* slot 6=cc */
    if (envelope.elementAt(6) instanceof String) {
      tmp=(String)envelope.elementAt(6);
      if (tmp.equals("NIL")==false) {
	throw new InternalError("Bogus value in the CC slot of the envelope");
      }
      cc=null;
    } else {
      cc=Util.parseAddresses((Vector)envelope.elementAt(6));    
    }
    /* slot 7=bcc */
    if (envelope.elementAt(7) instanceof String) {
      tmp=(String)envelope.elementAt(7);
      if (tmp.equals("NIL")==false) {
	throw new InternalError("Bogus value in the BCC slot of the envelope");
      }
      bcc=null;
    } else {
      bcc=Util.parseAddresses((Vector)envelope.elementAt(7));    
    }
    /* slot 8 is a string with the in-reply-to value */
    irt=(String)envelope.elementAt(8);
    /* NIL? */
    if (irt.equals("NIL")==true) {
      irt=null;
    }
    /* slot 9 is the message ID */
    mid=(String)envelope.elementAt(9);

    /* build the message object */
    message=new Message(flags,size,date,subject,from,sender,replyTo,to,cc,
                        bcc,irt,mid,m,msgno);
    
    /* BOGUSLY, I've somehow screwed up my recursion so the base */
    /* case is not clean... This is probably XXX */
    /* test for multipart at first ... */
    if ( body.elementAt(0) instanceof Vector == true) {
      bp=MIME.formMultipart(body,basePart,message,msgno);
    } else {
      /* first one is not a multipart */
      bp=MIME.structureToPart(body,basePart,1,message,msgno);
    }
    /* set the message body part of the head of the list of body parts */
    message.setBodyPart(bp);
    /* return the value to the caller */
    return message;
  }

  /**
   * This function handles tearing apart the responses to FETCH messages.
   * 
   * @param String s the string that was received from the server 
   * @param int n the message number we fetched
   * @exception InternalErr is thrown if end up with a FETCH reply from the server on  a message that is not in the cache. This is usually a programmer error.
   */
  void handleFetch(String s, int n) throws InternalErr {
    int index,size;
    Vector envelope,result;
    String tmp;
    Message message;
    StringTokenizer st=new StringTokenizer(s);
    
    /* find first ( */
    index=s.indexOf("(");
    /* parse! */
    result=Util.buildList(s,index);
    /* remove excess parsing data */
    result=(Vector)Util.removeParsingData(result);
    /* get the flags .. they are in slot 1 */
    flags=(Vector)result.elementAt(1);
    /* now, check to see if this is the result of STORE message and we  */
    /* just have the flags...  Note: if this is true, we'll only have the flags */
    /* and thus will have only two elements in the result */
    if (result.size()==2) {
      /* ok, got new flags vector find message */
      message=getMessage(n);
      /* if its not in the cache, we have a programmer error */
      if (message==null) {
	throw new InternalErr("Received FLAGS FETCH response on message not in cache! This is probably a programmer error!");
      }
      /* ok, we're cool, just set the flags */
      message.setFlags(flags);
      /* bail out, we are done */
      return;
    }
    /* the RFC 822 size is in slot 3 */
    tmp=(String)result.elementAt(3);
    try {
      size=Integer.valueOf(tmp,10).intValue();
    } catch (NumberFormatException e) {
      /* bad size! */
      throw new InternalErr("Unable to parse the number of bytes in the RFC822.SIZE slot");
    }
    /* envelope is in slot 5*/
    envelope=(Vector)result.elementAt(5);
    /*  slot 7 is the body part information */
    /* now we have all we need */
    message=Mailbox.createMessage(flags,envelope,(Vector)result.elementAt(7),
                                  size,this,"",n);
    /* put it in the cache */
    messageCache.put(new Integer(n),message);

  }
  /** 
   * This variable tells you how many messages have been added since the
   * last time you visited the box.
   */
  int recent;
  /** 
   * This variable tells you how many messages are in a mailbox. 
   */
  int exists;
  /**
   * This function returns the number of "recent" messages in a mailbox.
   * See description of recent variable for more info.
   *
   * @exception InternalErr is thrown if you call getRecent() and the mailbox has not seen a RECENT message yet (this should never happen)
   * @return number of messages added to box since last read
   */
  public int getRecent() throws InternalErr {
    /* sanity check */
    if (recent==-1) {
      throw new InternalErr("getRecent() called but recent count is -1!");
    }
    return recent;
  };
  /**
   * This function returns the size of the mailbox in messages.
   *
   * @exception InternalErr is thrown if you call getExists() and the mailbox hasn't yet seen an EXIST message from the server (this should never happen)
   * @return number of messages in a mailbox
   */
  public int getExists() throws InternalErr{
    /* sanity check */
    if (exists==-1) {
      throw new InternalErr("getExists() called but exists count is -1!");
    }
    return exists;
  };
  /**
   * This function gets the current list of flags on this mailbox.
   *
   * @return the current Vector of flags (they are strings in the vector)
   */
  public Vector getFlags() { return flags;};
  /**
   * This function gets the current list of flags that you can modify
   * permanently on this mailbox.  Note: this is an optional message from 
   * the server. It might not send it, so you must be able to handle the
   * case when this vector is empty. Further, if you open a mailbox read-
   * only this vector will be empty (as you can't modify that mailbox!).
   *
   * @return the current Vector of permanent flags (they are strings in the vector)
   */
  public Vector getPermanentFlags() { return permanentFlags;};
  /**
   * This variable holds a list of strings that you will see as the flags
   * on the messages from the server. 
   */
  Vector flags;
  /**
   * This variable holds a list of the permanent flags that you can modify
   * on this mailbox.
   */
  Vector permanentFlags;
  /**
   * Handle the PERMANENTFLAGS optional message from the server.
   *
   * @param s string with the server message in it
   */
  void handlePermanentFlags(String s){
    StringTokenizer st=new StringTokenizer(s,"(");
    String tmp;

    /* skip the part up to the first '(' note: this results in a removal */
    /* of the leading paren, so I put it back later */
    st.nextToken();
    /* now grab the part up to the end of the list (the final paren) */
    tmp=st.nextToken(")");
    /* add the inital paren back */
    tmp="(" + tmp;
    /* parse with the util code and put in the permanent flags slot */
    permanentFlags=Util.parseStringList(tmp);
  }
  /**
   * Fetch a mail message from the mailbox.  This is really a call to 
   * FETCH (FLAGS RFC822.SIZE ENVELOPE BODYSTRUCTURE), if you are 
   * looking at RFC 1730. Note: This call has the side effect of updating
   * this mailboxes cache if necessary. If not necessary, it does nothing.
   * Thus, call this whenever you want!
   * 
   * @param s string with a message number or message sequence in it (see RFC 1730 for how to form a sequence)
   * @exception InternalErr is propagated
   * @exception IMAPException  is propagated
   * @return boolean if we succeed or failed
   *
   */
  public boolean FETCH(String s) throws InternalErr, IMAPException {
    Vector result=new Vector();
    Integer i=Integer.valueOf(s,10);

    /* check the cache */
    if (messageCache.containsKey(i)==true) {
      return true;
    }
    /* send FETCH to server */
    getProtocol().getIO().roundTrip("FETCH " + s + " (FLAGS RFC822.SIZE ENVELOPE BODYSTRUCTURE)",result);
    /* dispatch the results */
    getProtocol().dispatchVector(result,false);
    
    /* see if we succeeded */
    return (Util.replyToken(result,"OK")==true); 
  }
  /**
   * Another version of FETCH() which takes an integer. This is just a wrapper around
   * the real FETCH. 
   * @param int n message number to fetch
   * @exception IMAPException is propagated
   * @exception InternalErr is propagated
   * @return true if the FETCH succeeded, false otherwise
   */
  public boolean FETCH(int n) throws IMAPException, InternalErr {
    Integer i=new Integer(n);
    return (FETCH(i.toString()));
  }
  /**
   * Get a message from the message cache. If you have not yet FETCHed
   * the message, then it won't be in the cache and you'll get null. You
   * will also get null if you ask for a message number that is out of
   * bounds.
   *
   * @param int n message number to retreive
   * @return the Message object corresponding to that message number
   */
  public Message getMessage(int n) {
    Integer i=new Integer(n);
    
    /* see if its valid.. */
    if (messageCache.containsKey(i)==true) {
      return (Message)messageCache.get(i);
    } 
    /* not in cache, so bail out */
    return null;
  }
  /** 
   * This variable holds the message cache. Its a mapping from message 
   * number to message object. 
   */
  Hashtable messageCache;
  /**
   * This is a function you can call when you want to flush the cache
   * by hand. It will be called automatically by the object itself at 
   * certain points in the protocol as necessary.
   */
  public void flushCache() {
    /* tell table flush itself...*/
    messageCache.clear();
  }
  /**
   * This is an internal use function. It allows a BodyPart object to
   * set itself up as the recipient of a FETCH message when it 
   * comes into this object. Note: This short circuits the usual
   * processing of FETCH messages, so this should only be used
   * while you are waiting for a response to getting the BodyPart's
   * data. Note: setting this to an object enables it, setting it to 
   * null disables.
   *
   */
  void setBodyPartReceiver(BodyPart bp) { receiver=bp;};
  /**
   * This variable holds the body part with the receiver
   */
  BodyPart receiver;
  /**
   * This an internal use function so other parts of the system 
   * can get the protocol object out of this mailbox.
   *
   * @return a Protocol object associated with this mailbox 
   */
  Protocol getProtocol() { return getConnection().getProtocol();};
  /**
   * This an internal use function so other parts of the system 
   * can get the Connection  object out of this mailbox.
   *
   * @return a Connection object associated with this mailbox 
   */
  Connection getConnection() { return connection;};
  /**
  * This function is called to actual delete all the messages in the mailbox with 
  * the \Deleted flag set.  As per RFC 1730, if there is an EXPUNGEd messages 
  * this will result in all messages with higher numbers automatically being
  * decremented by one. This happens BEFORE this call returns, so immediately
  * following this call, you should probably think about what the state of the
  * message cache might be. 
  *
   * @exception IMAPException is propagated
   * @exception InternalErr is propagated
  * @return bool true if the EXPUNGE call succeeded, false otherwise
  */
  public boolean EXPUNGE() throws InternalErr, IMAPException {
    Vector result=new Vector();
    
    /* send the command to the server */
    getProtocol().getIO().roundTrip("EXPUNGE",result);
    /* dispatch results */
    getProtocol().dispatchVector(result,false);
    /* return result to user */
    return (Util.replyToken(result,"OK")==true);
  }
  /**
   * This variable holds the link to our connection.
   */
  Connection connection;
  /**
   * CLOSE causes a mailbox to go away.  It updates the Connection object to 
   * reflect this. Note: If there are pending deletions (i.e. Messages marked with
   * the \Deleted flag) they get deleted as a result of this. However, you won't 
   * get any notification of this.  Note, if the CLOSE fails, then we don't tell
   * the Connection object to drop its Mailbox. 
   * 
   * @exception IMAPException is propagated
   * @exception InternalErr is propagated
   * @return true if the CLOSE succeeded, false otherwise
   */
  public boolean CLOSE() throws IMAPException, InternalErr{
    Vector result=new Vector();
    
    /* send the command to the server */
    getProtocol().getIO().roundTrip("CLOSE",result);
    /* dispatch results ... might be stuff besides EXPUNGE, eh? */
    getProtocol().dispatchVector(result,false);
    /* return result to user */
    if (Util.replyToken(result,"OK")==true) {
      /* we succeeded,  so tell the connection object we are gone */
      getConnection().mailboxClosed();
      return true;
    }
    /* we failed */
    return false;
  }
  /**
  * This function is called to get the server to (posibly) checkpoint its state,
  * if that is necesary. Note: This does not imply that you will get new EXISTS
  * messages. If you want that type of stuff, use NOOP.  Note: This  may not 
  * instanteous. For more info on this and the differences from NOOP, see RFC 1730.
  *
  * @exception IMAPException is propagated
  * @exception InternalErr is propagated
  * @return bool true if the CHECK call succeeded, false otherwise
  */
  public boolean CHECK() throws InternalErr, IMAPException {
    Vector result=new Vector();
    
    /* send the command to the server */
    getProtocol().getIO().roundTrip("CHECK",result);
    /* dispatch results */
    getProtocol().dispatchVector(result,false);
    /* return result to user */
    return (Util.replyToken(result,"OK")==true);
  }
  /** This function is called to COPY a message from one mailbox to another. 
   * This function has three possible return values, true, false, and null.  Note:
   * these are returned as the <BOLD>class</BOLD> Boolean, not simply booleans.
   * In other words be SURE to use the the equals() method not the == operator.
   * Example:
   * <PRE>
   * Boolean b;
   * Mailbox m;
   * ...
   * b=m.COPY(2,"~/somebox");
   * if (b==null) {
   * ... try to create the mailbox ... and retry if you want ...
   * } else {
   *    if (b.equals(new Boolean(true))==true) {
   *        ... it worked ...
   *    } else {
   *        ... failed miserably... creating it won't help...
   *    }
   * 
   * } 
   * </PRE>
   * If the return value is
   * true, the call succeded. If its false, the call failed and a attempting to create
   * the mailbox will not help. If its null (and you <BOLD>better</BOLD> check
   * for it) then this call failed, but using the CREATE command could create the
   * mailbox so that retrying this call would work.  Note: It seems that you might 
   * want to check for the null case first just to make sure you don't end up using
   * a null pointer if indeed it is null. 
   * 
   * @param int n message number
   * @param String box name of target mailbox
   * @exception InternalErr is propagated
   * @exception IMAPException is propagated
   * @return a Boolean which is either true, false, or null (see above)
   */
  public Boolean COPY(int n, String box) throws InternalErr, IMAPException{
    Vector result=new Vector();
    Integer i =new Integer(n);
    String tmp;

    /* send the command to the server */
    getProtocol().getIO().roundTrip("COPY " + i.toString() + " " + box,result);
    /* dispatch results */
    getProtocol().dispatchVector(result,false);
    /* return result to user */
    if (Util.replyToken(result,"OK")==true) {
      return (new Boolean(true));
    }
    /* we know it failed, should we try to create it */
    tmp=(String)result.elementAt(result.size()-1);
    /* does the last line contain the TRYCREATE business */
    if (tmp.indexOf("[TRYCREATE]")!=-1) {
      /* tell them what happened ... they can create it */
      return null;
    }
    /* total lose */
    return (new Boolean(false));
  }
  /** 
   * This function is called when we get notified that a message has been expunged.
   * @param String s the string that was received from the server 
   * @param int n the message number we expunged
   * @exception InternalErr is propagated
   */
  void handleExpunge(String s, int n) throws InternalErr {
    Integer i,j,key;
    /* ok, we need to muck around in the cache... */
    int num;
    /* temporary holding place for messages...note: they are not cast to their real */
    /* type for performance reasons */
    Object o;

    /* find a key as an integer */
    key=new Integer(n);
    /* sanity check */
    if (messageCache.containsKey(key)==false) {
      /* its there, clear it out */
      messageCache.remove(key);
    }
    /* ok, now look through all higher numbered messages and try to renumber them */
    for (num=n+1; num<=getExists(); ++num) {
      i=new Integer(num);
      /* is it in cache? */
      if (messageCache.containsKey(i)==true) {
	/* create a new message number that's one less. Note: this always works */
	/* because we are walking up the numbers so any lower numbered message */
	/* has already been renumbered by the time we reach here*/
	j=new Integer(num-1);
	/* remove old and insert new with old value */
	o=messageCache.get(i);
	messageCache.remove(i);
	messageCache.put(j,o);
      }
    }
    /* do we have any action to perform ? */
    if (expungeAction!=null) {
      expungeAction.run("EXPUNGE",key);
    }
  }
  /**
   * This function sets up a callback AsynchronousAction which gets invoked whenever
   * new mail is received. This is triggered by reception of the RECENT response in the
   * IMAP protocol.  Note:   when you first open a mailbox it may have new mail in it and
   * you will not get a call on this object. The reason is that the RECENT message to the
   * mailbox is generated <EM>before</EM> the mailbox object is visible outside the
   * Connection object it lives in. Note: When we call the action object you pass in 
   * to here, we supply an Integer (<BOLD>not an int</BOLD>) which has the value of
   * number of messages which are new.
   * 
   * @param AsynchronousAction a AsynchAction object to invoke on recent messages
   */
  public void setRecentAction(AsynchronousAction a) { recentAction=a;};
  /**
   * This variable holds the asynch. action we are using for the recent callback.
   */
  AsynchronousAction recentAction;
  /**
   * This variable holds the action which will get called when a mailbox gets the
   * EXPUNGE message. 
   */
  AsynchronousAction expungeAction;
  /**
   * This function sets the Aysnch.Action object which is associated with EXPUNGE
   * messages.  This indicates the messages are getting deleted by the
   * server. As soon as a message is deleted, all higher number messages
   * need to have their message numbers decremented by one. See RFC 1720 for
   * more information on this. Also, by the time you get the expunge message
   * the toolkit has already removed the message from its data structures so 
   * you can't FETCH() it or getData() it.  It also has already renumbered all 
   * messages with higher numbers if you try to access them. 
   * 
   * The parameter supplied to your function is an Integer
   *  (<BOLD> not an int </BOLD>) which is the message number  that has
   *  been expunged.
   * 
   * It would probably behoove programmers using this API to simply do all
   * message deletion processing in this callback.  Since you get one of these 
   * for each deletion, it just makes life easy.
   *
   * @param AsynchronousAction a AsynchAction object to invoke on expunge messages
   */
  public void setExpungeAction(AsynchronousAction a) { expungeAction=a;};
}
