/*
 * $Id: Message.java,v 1.4 1996/01/19 01:56:52 iansmith Exp $
 * $Author: iansmith $
 * $Log: Message.java,v $
 * Revision 1.4  1996/01/19  01:56:52  iansmith
 * Removed code that I had in there to automatically remove the \Recent
 * flag from the set of flags if you read a message. This confuses things
 * when combined with a future STORE since the server still has it marked
 * \Recent and you get that information in the response to your STORE.
 *
 * Documentation improvements in the javadoc comments.
 *
 * Revision 1.3  1996/01/15  18:51:01  iansmith
 * Added a function called nowSeen() which will allow other parts of the
 * system to mark this message as having been seen. This implicitly
 * results in removing the \Recent flag if present.
 *
 * Added the STORE() function to change the flags of a message.
 *
 * Added a function called setFlags() which allows the Mailbox object
 * to set the flags on the message. This is a result of a successful
 * STORE command.
 *
 * Revision 1.2  1996/01/10  22:01:20  iansmith
 * Added a couple of new slots to (and parameters to the constructor of)
 * the Message object. These are a reference back to the containing
 * mailbox and a message number. I needed both of these so when we
 * perform a FETCH() on a body part it can figure out what mailbox
 * its in and what message its in.
 *
 * Added supporting code to get and set the BodyPart() associated with
 * the first part of a message.
 *
 * Added a function get the mailbox of a message object... The BodyPart
 * objects need this.
 *
 * Added the getNumber() function which gives you the message number as
 * a string. This can get hairy if you are getting the number of a
 * message which is encapsulated inside another message.
 *
 * Revision 1.1  1996/01/08  06:01:17  iansmith
 * Initial revision
 *
 *
 */

package GIT.GVU.IMAP;
import java.lang.*;
import java.util.*;

/**
 * Class which represents a single message retrieved from an IMAP server.
 * 
 * Note: Some of the fields of this class may be evaluated lazily. Although,
 * this should not matter to the programmer, it may matter to the user.
 *
 * @version $Id: Message.java,v 1.4 1996/01/19 01:56:52 iansmith Exp $
 * @author Ian Smith
 */

public class Message {
  /**
   * This boolean controls whether or not the text of the message has 
   * been fetched yet. 
   */
  boolean bodyFetched;
  /**
   * Construct a message object.
   *
   * @param Vector f flags as a vector of strings
   * @param int s size of the message in bytes
   * @param String d RFC 822 style date
   * @param String subj subject of the message
   * @param Vector fr Vector of mail address specifying from line
   * @param Vector send Vector of mail address specifying sender line
   * @param Vector rt Vector of mail address specifying reply-to line
   * @param Vector t vector of mail addresses specifying to line
   * @param Vector c vector of mail addresses specifying cc line
   * @param Vector b vector of mail addresses specifying bcc line
   * @param String irt in-reply-to line
   * @param String mid message ID
   * @param Mailbox m mailbox associated with this Message
   * @param int msgno message number for this message
   *
   */
  public Message(Vector f, int s, String d, String subj, Vector fr,
                 Vector send, Vector rt, Vector t, Vector c, 
                 Vector b, String irt, String mid, Mailbox box, int msgno) {
    bodyFetched=false;
    /* assign the vars to the obvious slots */
    flags=f;
    size=s;
    date=d;
    subject=subj;
    from=fr;
    sender=send;
    replyTo=rt;
    to=t;
    cc=c;
    bcc=b;
    inReplyTo=irt;
    messageID=mid;
    mailbox=box;
    number=msgno;
  }
  /* 
   * locals that hold the obvious values for these slots in this object
  */
  Vector from;
  Vector sender;
  Vector replyTo;
  Vector to;
  Vector cc;
  Vector bcc;
  Vector flags;
  int size;
  String date;
  String inReplyTo;
  String subject;
  String messageID;
  BodyPart bodyPart;
  Mailbox mailbox;
  /**
   * Get the value of the from field. Note: This is usually going to be 
   * a Vector of size 1.
   *
   * @return Vector of the MailAddress objects which is the from field
   */
  public Vector getFrom() { return from;};
  /**
   * Get the value of the sender field. Note: This is usually going to be 
   * a Vector of size 1. Note: When this is not the same as the from field,
   * it indicates something like a mailing list or other type of repeater.
   *
   * @return Vector of the MailAddress objects which is the sender field
   */
  public Vector getSender() { return sender;};
  /**
   * Get the value of the reply-to field. Note: This is usually going to be 
   * a Vector of size 1.
   *
   * @return Vector of the MailAddress objects which is the reply-to field
   */
  public Vector getReplyTo() { return replyTo;};
  /**
   * Get the value of the to field. 
   *
   * @return Vector of the MailAddress objects which is the to field.
   */
  public Vector getTo() { return to;};
  /**
   * Get the value of the cc field. Note: This might be null.
   *
   * @return Vector of the MailAddress objects which is the cc field.
   */
  public Vector getCC() { return cc;};
  /**
   * Get the value of the bcc field. Note: This might be null.
   *
   * @return Vector of the MailAddress objects which is the bcc field.
   */
  public Vector getBCC() { return bcc;};
  /**
   * Get the flags that are set for this message. This is a vector of
   * strings like \Seen and \Answered..  Note:This will be null if 
   * the message is embedded in another message. 
   *
   * @return a vector of strings with the value of the flags
   */
  public Vector getFlags() { return flags;};
  /**
   * Get the size of the message in bytes.
   * 
   * @return an integer with the size of the message in bytes
   */
  public int getSize() { return size;};
  /**
   * Get the subject line of this message.
   *
   * @return a string with the subject of the message.
   */
  public String getSubject() { return subject;};
  /**
   * Get the Message ID of this mesage.
   *
   * @return a string with this message's message ID
   */
  public String getMessageID() { return messageID;};
  /**
   * Get the In-Reply-To line of this mesage. Note: This usually has
   * a message ID in it so you can do threading. Note: This might be
   * null if its not a reply.
   *
   * @return a string with the in-reply-to line of the message
   */
  public String getInReplyTo() { return inReplyTo;};
  /**
   * Get the date of this message. This is returned as a string, which 
   * is probably bogus. XXX Fix me! Should return a date object XXX 
   * 
   * @return a string with the date of the message
   */
  public String getDate() { return date;};
  /**
   * Get the BodyPart object that corresponds to the main  (first) portion
   * of this message.
   *
   * @return BodyPart object which is the body for this message.
   */
  public BodyPart getBodyPart() { return bodyPart;};
  /**
   * Set the body part associated with this message. This is basically
   * the head of the list for multipart messages. 
   */
  public void setBodyPart(BodyPart bp) { bodyPart=bp;};
  /**
   * Get the mailbox this message is in.
   * 
   * @return mailbox this message is in
   */
  public Mailbox getMailbox() { return mailbox;};
  /**
   * Printable version of the message number. Note: This may be a little
   * deceptive with messages that are encapsulated in other messages
   * via MIME. This will reflect the message number that you need to
   * FETCH to get to this message.  Thus the internal message and the
   * "normal" message will have the same number.
   * 
   * @return a string with the message number in it
   */
  String getNumber() { return (new Integer(number)).toString();};
  /**
   * This is where we hold this message's message #
   */
  int number;
  /**
   * Update this messages local cache of flags to include the \Seen  flag.
   */
  void nowSeen() {
    /* check to see if we have already seen it? */
    if (Util.stringInVector(flags,"\\Seen")==false) {
      flags.addElement("\\Seen");
    }
  }
  /**
   * This function causes the flag set of a message to change. This is 
   * most commonly used to mark messages for deletion. 
   * 
   * @param String t  This is a string whose value must be "+", "-", or "".  If its + we will add the flag to this message, - means remove a flag, and empty means set the flags to this value.
   * @param Object o. If you are adding or subtracting flags this should be a string. If you are setting the flags, this should be a vector of strings.
   * @exception IMAPException is propagated
   * @exception InternalErr is thrown if you call STORE with a bad combination of parameters
   * @return true if the command succeeded.
   */
  public boolean STORE(String t,Object o) throws InternalErr, IMAPException {
    String flagString;
    int i;
    Vector v,result=new Vector();

    /* are we setting the flags */
    if (t.equals("")==true) {
      /* make sure its a vector: sanity check */
      if (o instanceof Vector==false) {
	throw new InternalErr("Bad use of STORE in an attempt to set the flags");
      }
      /* set vector */
      v=(Vector)o;
      /* loop over all flags, building a string */
      flagString="(";
      for (i=0; i<v.size(); ++i) {
	flagString+= (String)v.elementAt(i);
	/* do we need a space to separate ?*/
	if (i!=v.size()-1) {
	  flagString+=" ";
	}
      }
      flagString+=")";
    } else {
      /*  this is either + or - right? */
      if ((t.equals("+")==false) && (t.equals("-")==false)) {
	throw new InternalErr("Bad parameter value for type of STORE command wanted");
      }
      /* ok, then lets add or subtract a flag */
      flagString="(" + (String)o + ")";
    }
    /* send the command */
    mailbox.getProtocol().getIO().roundTrip("STORE " + getNumber() + " " + t + 
                                            "FLAGS " + flagString,result);
    /* issue the results */
    mailbox.getProtocol().dispatchVector(result,false);
    /* did it succeed */
    return (Util.replyToken(result,"OK")==true);
  }
  /** 
   * This function is here to allow the Mailbox to set the flags on a message when it 
   * gets a set of flags as a result of STORE command. 
   * 
   * @param Vector v the vector of strings with the flags in it
   */
  void setFlags(Vector v) {
    flags=v;
  }
}
